#include "TerrainModel.h"

bool TerrainModel::Initialize(char* filename)
{
	LoadRawHeightMap(filename);

	SetTerrainCoordinates();

	CalculateNormals();

	BuildTerrainModel();
	
	InitializeQuadTree();

	InitializeTerrainThreading();

	return true;
}

bool TerrainModel::LoadHeightMap(char* filename)
{
	FILE* filePtr;
	int error;
	unsigned int count;
	BITMAPFILEHEADER bitmapFileHeader;
	BITMAPINFOHEADER bitmapInfoHeader;
	int imageSize, k, index;
	unsigned char* bitmapImage;
	unsigned char height;
	int terrainWidth, terrainHeight;
	float* heightMapData;

	// Open the height map file in binary.
	error = fopen_s(&filePtr, filename, "rb");
	if (error != 0)
	{
		return false;
	}

	// Read in the file header.
	count = fread(&bitmapFileHeader, sizeof(BITMAPFILEHEADER), 1, filePtr);
	if (count != 1)
	{
		return false;
	}

	// Read in the bitmap info header.
	count = fread(&bitmapInfoHeader, sizeof(BITMAPINFOHEADER), 1, filePtr);
	if (count != 1)
	{
		return false;
	}

	terrainWidth = bitmapInfoHeader.biWidth;
	terrainHeight = bitmapInfoHeader.biHeight;

	m_terrainWidth = terrainWidth;
	m_terrainHeight = terrainHeight;

	imageSize = terrainWidth * terrainHeight * 3;

	bitmapImage = new unsigned char[imageSize];
	if (!bitmapImage)
	{
		return false;
	}

	//Move to the beginning of the image data
	fseek(filePtr, bitmapFileHeader.bfOffBits, SEEK_SET);

	//Read in the bitmap data
	count = fread(bitmapImage, 1, imageSize, filePtr);
	if (count != imageSize)
	{
		return false;
	}

	error = fclose(filePtr);
	if (error != 0)
	{
		return false;
	}

	heightMapData = new float[terrainWidth * terrainHeight];
	if (!heightMapData)
	{
		return false;
	}

	k = 0;

	for (int j = 0; j < terrainHeight; j++)
	{
		for (int i = 0; i < terrainWidth; i++)
		{
			height = bitmapImage[k];

			index = (terrainHeight * j) + i;

			heightMapData[index] = (float)height;

			//heightMapData[index] /= 256.0f;

			k += 3;
		}
	}

	for (int i = 0; i < m_vertexCount; i++)
	{
		int map_index;
		float mapX, mapZ;

		if (m_vertices[i].x == 0.5f)
		{
			mapX = 0.5f - (1.0f / m_terrainWidth);
		}
		else

		{
			mapX = m_vertices[i].x;
		}

		if (m_vertices[i].z == 0.5f)
		{
			mapZ = 0.5f - (1.0f / m_terrainHeight);
		}
		else

		{
			mapZ = m_vertices[i].z;
		}

		map_index = (((mapX + 0.5f) / (1.0f / (float)m_terrainWidth)) * (float)m_terrainHeight) + ((mapZ + 0.5f) / (1.0f / (float)m_terrainHeight));
		m_vertices[i].y = (heightMapData[map_index] / 256.0f);
	}


	delete[] bitmapImage;
	bitmapImage = nullptr;

	return true;
}

void TerrainModel::Shutdown()
{
	if (m_parentNode)
	{
		ReleaseNode(m_parentNode);
		delete m_parentNode;
		m_parentNode = nullptr;
	}

	elapsed = true;

	for (int i = 0; i < THREAD_COUNT; i++)
	{
		m_taskChannel.release_all();
		m_threadList[i].join();
	}
}

void TerrainModel::InitializeTerrainThreading()
{
	elapsed = false;
	task_count = 0;

	for (int i = 0; i < THREAD_COUNT; i++)
	{
		m_threadList[i] = thread(&TerrainModel::ManipulateVertices, this, &m_taskChannel, &elapsed, &task_count, i);
	}
}

void TerrainModel::Render()
{
	glBindTexture(GL_TEXTURE_2D, GetTexture());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	RenderNode(m_parentNode);
}

bool TerrainModel::LoadRawHeightMap(char* filename)
{
	int error, index;
	bool result;
	FILE* filePtr;
	unsigned long long imageSize, count;
	unsigned short* rawImage;
	float* heightMapData;

	m_terrainWidth = 256;
	m_terrainHeight = 256;

	m_heightMap = new VertexType[m_terrainWidth * m_terrainHeight];
	if (!m_heightMap)
	{
		return false;
	}

	imageSize = m_terrainWidth * m_terrainHeight;

	error = fopen_s(&filePtr, filename, "rb");
	if (error != 0)
	{
		return false;
	}

	rawImage = new unsigned short[imageSize];
	if (!rawImage)
	{
		return false;
	}

	count = fread(rawImage, sizeof(unsigned short), imageSize, filePtr);
	if (count != imageSize)
	{
		return false;
	}

	error = fclose(filePtr);
	if (error != 0)
	{
		return false;
	}

	for (int j = 0; j < m_terrainHeight; j++)
	{
		for (int i = 0; i < m_terrainWidth; i++)
		{
			index = (m_terrainWidth * j) + i;

			m_heightMap[index].y = (float)rawImage[index];
		}
	}

	delete[] rawImage;
	rawImage = nullptr;

	return true;
}

void TerrainModel::SetTerrainCoordinates()
{
	int index;

	for (int j = 0; j < m_terrainHeight; j++)
	{
		for (int i = 0; i < m_terrainWidth; i++)
		{
			index = (m_terrainWidth * j) + i;
			
			m_heightMap[index].x = (float)i;
			m_heightMap[index].z = -(float)j;

			//centre the heightmap around the origin
			m_heightMap[index].x -= (float)(m_terrainWidth - 1) / 2;
			m_heightMap[index].z += (float)(m_terrainHeight - 1)/2;

			m_heightMap[index].y /= 300.0f;
		}
	}
}

bool TerrainModel::BuildTerrainModel()
{
	int index, index1, index2, index3, index4;

	m_vertexCount = (m_terrainHeight - 1) *(m_terrainWidth - 1) * 6;

	m_indices = new unsigned long[m_vertexCount];
	if (!m_indices)
	{
		return false;
	}

	index = 0;

	for (int j = 0; j < (m_terrainHeight - 1); j++)
	{
		for (int i = 0; i < (m_terrainWidth - 1); i++)
		{
			index1 = (m_terrainWidth * j) + i;
			index2 = (m_terrainWidth* j) + (i + 1);
			index3 = (m_terrainWidth * (j + 1)) + i;
			index4 = (m_terrainWidth * (j + 1)) + (i + 1);

			m_indices[index] = index1;
			m_heightMap[index1].tu = 0.0f;
			m_heightMap[index1].tv = 0.0f;
			index++;

			m_indices[index] = index2;
			m_heightMap[index2].tu = 1.0f;
			m_heightMap[index2].tv = 0.0f;
			index++;

			m_indices[index] = index4;
			m_heightMap[index4].tu = 1.0f;
			m_heightMap[index4].tv = 1.0f;
			index++;

			m_indices[index] = index1;
			m_heightMap[index1].tu = 0.0f;
			m_heightMap[index1].tv = 0.0f;
			index++;

			m_indices[index] = index4;
			m_heightMap[index4].tu = 1.0f;
			m_heightMap[index4].tv = 1.0f;
			index++;

			m_indices[index] = index3;
			m_heightMap[index3].tu = 0.0f;
			m_heightMap[index3].tv = 1.0f;
			index++;

		}
	}

	return true;
}

void TerrainModel::InitializeQuadTree()
{
	int vertexCount;
	float centerX, centerZ, width;

	vertexCount = m_vertexCount;

	m_triangleCount = vertexCount / 3;

	CalculateMeshDimensions(vertexCount, centerX, centerZ, width);

	m_parentNode = new NodeType;

	CreateTreeNode(m_parentNode, centerX, centerZ, width);
}

void TerrainModel::CalculateMeshDimensions(int vertexCount, float& centerX, float& centerZ, float& meshWidth)
{
	float maxWidth, maxDepth, minWidth, minDepth, width, depth, maxX, maxZ;

	centerX = 0.0f;
	centerZ = 0.0f;

	for (int i = 0; i < vertexCount; i++)
	{
		centerX += m_heightMap[m_indices[i]].x;
		centerZ += m_heightMap[m_indices[i]].z;
	}

	centerX = centerX / (float)vertexCount;
	centerZ = centerZ / (float)vertexCount;

	maxWidth = 0.0f;
	maxDepth = 0.0f;

	minWidth = fabsf(m_heightMap[m_indices[0]].x - centerX);
	minDepth = fabsf(m_heightMap[m_indices[0]].z - centerZ);

	for (int i = 0; i < vertexCount; i++)
	{
		width = fabsf(m_heightMap[m_indices[i]].x - centerX);
		depth = fabsf(m_heightMap[m_indices[i]].z - centerZ);

		if (width > maxWidth){ maxWidth = width; }
		if (depth > maxDepth){ maxDepth = depth; }
		if (width < minWidth){ minWidth = width; }
		if (depth < minDepth){ minDepth = depth; }
	}

	maxX = (float)std::max(fabs(minWidth), fabs(maxWidth));
	maxZ = (float)std::max(fabs(minDepth), fabs(maxDepth));

	meshWidth = max(maxX, maxZ) * 2.0f;
}

void TerrainModel::CreateTreeNode(NodeType* node, float positionX, float positionZ, float width)
{
	int numTriangles, count, vertexCount, index, vertexIndex;
	float offsetX, offsetZ;
	unsigned long* indices;
	bool result;

	node->positionX = positionX;
	node->positionZ = positionZ;
	node->width = width;

	node->triangleCount = 0;

	node->indices = nullptr;

	node->nodes[0] = nullptr;
	node->nodes[1] = nullptr;
	node->nodes[2] = nullptr;
	node->nodes[3] = nullptr;

	numTriangles = CountTriangles(positionX, positionZ, width);

	if (numTriangles == 0)
	{
		return;
	}

	if (numTriangles > MAX_TRIANGLES)
	{
		for (int i = 0; i < 4; i++)
		{
			offsetX = (((i % 2) < 1) ? -1.0f : 1.0f) * (width / 4.0f);
			offsetZ = (((i % 4) < 2) ? -1.0f : 1.0f) * (width / 4.0f);

			count = CountTriangles((positionX + offsetX), (positionZ + offsetZ), (width / 2.0f));
			if (count > 0)
			{
				node->nodes[i] = new NodeType;

				CreateTreeNode(node->nodes[i], (positionX + offsetX), (positionZ + offsetZ), (width/2.0f));
			}
		}

		return;
	}

	node->triangleCount = numTriangles;

	vertexCount = numTriangles * 3;

	node->indices = new unsigned long[vertexCount];

	thread threads[THREAD_COUNT];
	NodeThreadArgs threadArgs[THREAD_COUNT];

	index = 0;
	
	for (int i = 0; i < THREAD_COUNT; i++)
	{
		int start_index = ((float)m_triangleCount / (float)THREAD_COUNT) * (float)i;
		int end_index = ((float)m_triangleCount / (float)THREAD_COUNT) * (float)(i + 1);

		threadArgs[i] = NodeThreadArgs{ start_index, end_index, positionX, positionZ, width, node, &index};

		threads[i] = thread(&TerrainModel::CalculateNodeThreaded, this, threadArgs[i]);
	}

	for (int i = 0; i < THREAD_COUNT; i++)
	{
		threads[i].join();
	}
	
}

void TerrainModel::CalculateNodeThreaded(NodeThreadArgs args)
{
	bool result;
	int vertexIndex;

	for (int i = args.start_index; i < args.end_index; i++)
	{
		result = IsTriangleContained(i, args.positionX, args.positionZ, args.width);
		if (result)
		{
			vertexIndex = i * 3;

			std::lock_guard<mutex> lock(count_mutex);
			args.node->indices[*(args.index)] = m_indices[vertexIndex];
			*(args.index) += 1;
			
			vertexIndex++;
			args.node->indices[*(args.index)] = m_indices[vertexIndex];
			*(args.index) += 1;

			vertexIndex++;
			args.node->indices[*(args.index)] = m_indices[vertexIndex];
			*(args.index) += 1;
		}
	}
}

int TerrainModel::CountTriangles(float positionX, float positionZ, float width)
{
	int tri_count;
	bool result;

	thread threads[THREAD_COUNT];
	TriangleThreadArgs threadArgs[THREAD_COUNT];

	tri_count = 0;
	
	for (int i = 0; i < THREAD_COUNT; i++)
	{
		int start_index = ((float)m_triangleCount / (float)THREAD_COUNT) * (float)i;
		int end_index = ((float)m_triangleCount / (float)THREAD_COUNT) * (float)(i+1);

		threadArgs[i] = TriangleThreadArgs{ start_index, end_index, positionX, positionZ, width, &tri_count };

		threads[i] = thread(&TerrainModel::IsTriangleContainedThreaded, this, threadArgs[i]);
	}

	for (int i = 0; i < THREAD_COUNT; i++)
	{
		threads[i].join();
	}
	
	return tri_count;
}

bool TerrainModel::IsTriangleContained(int index, float positionX, float positionZ, float width)
{
	float radius;
	int vertexIndex;
	unsigned long heightMapIndex;
	float x1, z1, x2, z2, x3, z3;
	float minimumX, maximumX, minimumZ, maximumZ;

	radius = width / 2.0f;

	vertexIndex = index * 3;

	heightMapIndex = m_indices[vertexIndex];
	x1 = m_heightMap[heightMapIndex].x;
	z1 = m_heightMap[heightMapIndex].z;
	vertexIndex++;

	heightMapIndex = m_indices[vertexIndex];
	x2 = m_heightMap[heightMapIndex].x;
	z2 = m_heightMap[heightMapIndex].z;
	vertexIndex++;

	heightMapIndex = m_indices[vertexIndex];
	x3 = m_heightMap[heightMapIndex].x;
	z3 = m_heightMap[heightMapIndex].z;
	vertexIndex++;

	minimumX = min(x1, min(x2, x3));
	if (minimumX > (positionX + radius))
	{
		return false;
	}

	maximumX = max(x1, max(x2, x3));
	if (maximumX < (positionX - radius))
	{
		return false;
	}

	minimumZ = min(z1, min(z2, z3));
	if (minimumZ > (positionZ + radius))
	{
		return false;
	}

	maximumZ = max(z1, max(z2, z3));
	if (maximumZ < (positionZ - radius))
	{
		return false;
	}

	return true;
}

void TerrainModel::IsTriangleContainedThreaded(TriangleThreadArgs args)
{
	float radius;
	int vertexIndex;
	unsigned long index;
	float x1, z1, x2, z2, x3, z3;
	float minimumX, maximumX, minimumZ, maximumZ;

	radius = args.width / 2.0f;

	for (int i = args.start_index; i < args.end_index; i++)
	{
		vertexIndex = i * 3;
		
		index = m_indices[vertexIndex];
		x1 = m_heightMap[index].x;
		z1 = m_heightMap[index].z;
		vertexIndex++;

		index = m_indices[vertexIndex];
		x2 = m_heightMap[index].x;
		z2 = m_heightMap[index].z;
		vertexIndex++;

		index = m_indices[vertexIndex];
		x3 = m_heightMap[index].x;
		z3 = m_heightMap[index].z;
		vertexIndex++;

		minimumX = min(x1, min(x2, x3));
		if (minimumX > (args.positionX + radius))
		{
			continue;
		}

		maximumX = max(x1, max(x2, x3));
		if (maximumX < (args.positionX - radius))
		{
			continue;
		}

		minimumZ = min(z1, min(z2, z3));
		if (minimumZ >(args.positionZ + radius))
		{
			continue;
		}

		maximumZ = max(z1, max(z2, z3));
		if (maximumZ < (args.positionZ - radius))
		{
			continue;
		}

		std::lock_guard<mutex> lock(count_mutex);
		*(args.count) += 1;
	}
}

void TerrainModel::ReleaseNode(NodeType* node)
{
	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i] != nullptr)
		{
			ReleaseNode(node->nodes[i]);
		}
	}

	if (node->indices)
	{
		delete[] node->indices;
		node->indices = nullptr;
	}

	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i])
		{
			delete node->nodes[i];
			node->nodes[i] = nullptr;
		}
	}
}

void TerrainModel::RenderNode(NodeType* node)
{
	int count;
	float x, y, z;

	count = 0;
	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i] != nullptr)
		{
			count++;
			RenderNode(node->nodes[i]);
		}
	}

	if (count != 0)
	{
		return;
	}

	glPushMatrix();
	GetScaling(x, y, z);
	glScalef(x, y, z);
	GetTranslation(x, y, z);
	glTranslatef(x, y, z);
	GetRotation(x, y, z);
	glRotatef(z, 0, 0, 1);
	glRotatef(y, 0, 1, 0);
	glRotatef(x, 1, 0, 0);

	glBegin(GL_TRIANGLES);


	for (int i = 0; i < (node->triangleCount * 3); i++)
	{
		unsigned long index = node->indices[i];
		glNormal3f(m_heightMap[index].nx, m_heightMap[index].ny, m_heightMap[index].nz);
		glTexCoord2f(m_heightMap[index].tu, m_heightMap[index].tv);
		glVertex3f(m_heightMap[index].x, m_heightMap[index].y, m_heightMap[index].z);
	}

	glEnd();
	glPopMatrix();
}

void TerrainModel::ManipulateTerrain(Vector3 manipulation_origin, float manipulation_radius, float manipulation_intensity, float attenuation_amount, AttenuationType attenuation_type)
{
	float left, right, top, bottom;
	NodeType* current_node = nullptr;

	FindNode(m_parentNode, manipulation_radius, manipulation_origin.x, manipulation_origin.z);

	left = m_nodeList.back()->positionX + (m_nodeList.back()->width / 2.0f);
	right = m_nodeList.back()->positionX - (m_nodeList.back()->width / 2.0f);
	top = m_nodeList.back()->positionZ + (m_nodeList.back()->width / 2.0f);
	bottom = m_nodeList.back()->positionZ - (m_nodeList.back()->width / 2.0f);

	float x = manipulation_origin.x;
	float z = manipulation_origin.z;

	if (x + manipulation_radius >= right)
	{
	FindNode(m_parentNode, 0.0f, x + manipulation_radius, z);
	}

	if (x - manipulation_radius <= left)
	{
	FindNode(m_parentNode, 0.0f, x - manipulation_radius, z);
	}

	if (z + manipulation_radius >= top)
	{
	FindNode(m_parentNode, 0.0f, x,  z + manipulation_radius);
	}

	if (z - manipulation_radius <= bottom)
	{
	FindNode(m_parentNode, 0.0f, x, z - manipulation_radius);
	}

	CheckVertexArgs check_args;

	the_clock::time_point begin = the_clock::now();

	while (!m_nodeList.empty())
	{
		current_node = m_nodeList.back();
		m_nodeList.pop_back();

		int start_index = 0;
		int end_index = ((float)(current_node->triangleCount * 3));

		check_args = CheckVertexArgs{ manipulation_origin, manipulation_radius, manipulation_intensity, attenuation_amount, attenuation_type, current_node, start_index, end_index };

		CheckVertices(check_args);
	}

	std::sort(m_taskList.begin(), m_taskList.end());

	std::unique(m_taskList.begin(), m_taskList.end());

	the_clock::time_point finish = the_clock::now();

	auto full_time = duration_cast<milliseconds>(finish - begin).count();

	std::ostringstream os_;
	os_ << "Creating the task list took " << full_time << " ms." << endl;
	OutputDebugString(os_.str().c_str());

	DataSnapshotEntry snapshot;

	snapshot.task_list_time = full_time;
	m_snapshotList.push_back(snapshot);
}

bool TerrainModel::SameData(const TaskListEntry& task1, const TaskListEntry& task2)
{
	return ((task1.index == task2.index) && (task1.frames_since_creation == task2.frames_since_creation));
}
void TerrainModel::CheckVertices(CheckVertexArgs args)
{
	float distance = 0.0f;
	Vector3 vertex_position;
	unsigned long index;
	bool index_found;
	TaskListEntry entry;

	for (int i = args.start_index; i < args.end_index; i++)
	{
		index_found = false;

		index = args.node->indices[i];

		vertex_position = Vector3(m_heightMap[index].x, m_heightMap[index].y, m_heightMap[index].z);
		distance = (vertex_position - args.origin_position).length();
		if (distance <= args.radius)
		{
			entry = TaskListEntry{ index, distance, args.manipulation_intensity, args.attenuation_amount, args.attenuation_type, 0 };
			//std::lock_guard<mutex> lock(task_list_mutex);
			m_taskList.push_back(entry);
			
		}
	}
}

void TerrainModel::Frame()
{
	TaskListEntry task;
	bool re_normals = false;

	elapsed = false;

	if (!m_taskList.empty() || !m_taskChannel.channel_empty())
	{
		re_normals = true;
	}

	if (re_normals)
	{
		task_count = 0;

		the_clock::time_point begin = the_clock::now();

		while (!m_taskList.empty())
		{
			task = m_taskList.back();
			m_taskList.pop_back();
			m_taskChannel.write(task);

			the_clock::time_point current_time = the_clock::now();
			auto full_time = duration_cast<milliseconds>(current_time - begin).count();

			if (full_time >= MAX_MANIPULATION_TIME)
				break;
		}

		std::ostringstream os_;
		os_ << "Processed " << task_count << " tasks." << endl;
		OutputDebugString(os_.str().c_str());

		m_snapshotList.back().tasks_processed.push_back(task_count);
	}

	for (auto task : m_taskList)
	{
		task.frames_since_creation++;
	}

	if (re_normals)
		CalculateNormals();
}

void TerrainModel::ManipulateVertices(channel<TaskListEntry>* task_channel, atomic_bool* elapsed, int* task_count, int thread_num)
{
	TaskListEntry task;

	float heightChange = 0.0f;

	while (elapsed->load() == false)
	{
		if (elapsed->load() == false)
		{
			task_channel->read(task);
		}

		if (elapsed->load() == true)
		{
			return;
		}

		switch (task.attenuation_type)
		{
		case AttenuationType::CONSTANT:
			heightChange = task.manipulation_intensity;
			break;
		case AttenuationType::LINEAR:
			heightChange = task.manipulation_intensity * (1.0f / (task.distance * task.attenuation_amount));
			break;
		case AttenuationType::QUADRATIC:
			heightChange = task.manipulation_intensity * (1.0f / (pow(task.distance, 2) * task.attenuation_amount));
			break;
		default:
			heightChange = 0.0f;
			break;
		}

		if (task.index < (m_triangleCount * 6) && task.index > 0)
			m_heightMap[task.index].y += heightChange;

		
		lock_guard<mutex> lock(task_count_mutex);
		(*task_count) += 1;
	}
}

void TerrainModel::WriteSnapshot(char* filename)
{
	int task_count = 0;
	time_t now = time(0);
	char time[30];
	std::string time_s;
	char file_string[5];
	char filename_string[30];
	ctime_s(time, sizeof time, &now);
	time_s = time;
	replace(time_s.begin(), time_s.end(), '\n', ' ');
	replace(time_s.begin(), time_s.end(), ':', '-');
	std::ofstream snapshot_file;
	snapshot_file.open(time_s, ios_base::app);
	if (snapshot_file.fail())
	{
		return;
	}
	snapshot_file << "Threads used: " << THREAD_COUNT << endl;
	while (m_snapshotList.size() > 0)
	{
		snapshot_file << "Processing Time: " << m_snapshotList.back().task_list_time << "ms" << endl;
		snapshot_file << "Frames Taken To Process: " << m_snapshotList.back().tasks_processed.size() << endl;
		for (int i = 0; i < m_snapshotList.back().tasks_processed.size(); i++)
		{
			snapshot_file << "Task Processed: " << m_snapshotList.back().tasks_processed[i] << endl;
			task_count += m_snapshotList.back().tasks_processed[i];
		}
		snapshot_file << "Total Tasks Processed: " << task_count << endl;
		snapshot_file << endl;

		m_snapshotList.pop_back();
	}
	snapshot_file.close();
}

bool TerrainModel::GetHeightAtPosition(float positionX, float positionZ, float& height)
{
	float meshMinX, meshMaxX, meshMinZ, meshMaxZ;

	meshMinX = (m_parentNode->positionX - (m_parentNode->width / 2.0f)) * m_scale.x;
	meshMaxX = (m_parentNode->positionX + (m_parentNode->width / 2.0f)) * m_scale.z;

	meshMinZ = (m_parentNode->positionZ - (m_parentNode->width / 2.0f)) * m_scale.x;
	meshMaxZ =( m_parentNode->positionZ + (m_parentNode->width / 2.0f)) * m_scale.z;


	if ((positionX <meshMinX) || (positionX > meshMaxX) || (positionZ < meshMinZ) || (positionZ > meshMaxZ))
	{
		return false;
	}

	FindNodeHeight(m_parentNode, positionX, positionZ, height);

	return true;
}


bool TerrainModel::GetNormalAtPosition(float positionX, float positionZ, Vector3& normal)
{
	float meshMinX, meshMaxX, meshMinZ, meshMaxZ;

	meshMinX = (m_parentNode->positionX - (m_parentNode->width / 2.0f)) * m_scale.x;
	meshMaxX = (m_parentNode->positionX + (m_parentNode->width / 2.0f)) * m_scale.z;

	meshMinZ = (m_parentNode->positionZ - (m_parentNode->width / 2.0f)) * m_scale.x;
	meshMaxZ = (m_parentNode->positionZ + (m_parentNode->width / 2.0f)) * m_scale.z;


	if ((positionX <meshMinX) || (positionX > meshMaxX) || (positionZ < meshMinZ) || (positionZ > meshMaxZ))
	{
		return false;
	}

	FindNodeNormal(m_parentNode, positionX, positionZ, normal);

	return true;
}

void TerrainModel::FindNode(NodeType* parent_node, float radius, float x, float z)
{
	float xMin, xMax, zMin, zMax;
	int count, index, vertexIndex;
	float vertex1[3], vertex2[3], vertex3[3];
	bool foundHeight;
	

	xMin = (parent_node->positionX - (parent_node->width / 2.0f)) * m_scale.x;
	xMax = (parent_node->positionX + (parent_node->width / 2.0f)) * m_scale.x;

	zMin = (parent_node->positionZ - (parent_node->width / 2.0f)) * m_scale.z;
	zMax = (parent_node->positionZ + (parent_node->width / 2.0f)) * m_scale.z;

	if ((x<xMin) || (x>xMax) || (z<zMin) || (z>zMax))
	{
		return;
	}

	count = 0;

	for (int i = 0; i < 4; i++)
	{
		if (parent_node->nodes[i] != nullptr)
		{
			count++;
			FindNode(parent_node->nodes[i], radius, x, z);
		}
	}

	if (count > 0)
	{
		return;
	}

	
	m_nodeList.push_back(parent_node);
}


void TerrainModel::FindNodeHeight(NodeType* node, float x, float z, float& height)
{
	float xMin, xMax, zMin, zMax;
	int count, index, vertexIndex;
	float vertex1[3], vertex2[3], vertex3[3];
	bool foundHeight;

	xMin = (node->positionX - (node->width / 2.0f)) * m_scale.x;
	xMax = (node->positionX + (node->width / 2.0f)) * m_scale.x;

	zMin = (node->positionZ - (node->width / 2.0f)) * m_scale.z;
	zMax = (node->positionZ + (node->width / 2.0f)) * m_scale.z;

	if ((x<xMin) || (x>xMax) || (z<zMin) || (z>zMax))
	{
		return;
	}

	count = 0;

	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i] != nullptr)
		{
			count++;
			FindNodeHeight(node->nodes[i], x, z, height);
		}
	}

	if (count > 0)
	{
		return;
	}

	for (int i = 0; i < node->triangleCount; i++)
	{
		index = i * 3;
		vertexIndex = node->indices[index];
		vertex1[0] = m_heightMap[vertexIndex].x;
		vertex1[1] = m_heightMap[vertexIndex].y;
		vertex1[2] = m_heightMap[vertexIndex].z;

		index++;
		vertexIndex = node->indices[index];
		vertex2[0] = m_heightMap[vertexIndex].x;
		vertex2[1] = m_heightMap[vertexIndex].y;
		vertex2[2] = m_heightMap[vertexIndex].z;

		index++;
		vertexIndex = node->indices[index];
		vertex3[0] = m_heightMap[vertexIndex].x;
		vertex3[1] = m_heightMap[vertexIndex].y;
		vertex3[2] = m_heightMap[vertexIndex].z;

		foundHeight = CheckHeightOfTriangle(x, z, height, vertex1, vertex2, vertex3);

		if (foundHeight)
		{
			return;
		}
	}
}

void TerrainModel::FindNodeNormal(NodeType* node, float x, float z, Vector3& normal)
{
	float xMin, xMax, zMin, zMax;
	int count, index, vertexIndex;
	float vertex1[3], vertex2[3], vertex3[3];
	bool foundHeight;
	float height;

	xMin = (node->positionX - (node->width / 2.0f)) * m_scale.x;
	xMax = (node->positionX + (node->width / 2.0f)) * m_scale.x;

	zMin = (node->positionZ - (node->width / 2.0f)) * m_scale.z;
	zMax = (node->positionZ + (node->width / 2.0f)) * m_scale.z;

	if ((x<xMin) || (x>xMax) || (z<zMin) || (z>zMax))
	{
		return;
	}

	count = 0;

	for (int i = 0; i < 4; i++)
	{
		if (node->nodes[i] != nullptr)
		{
			count++;
			FindNodeNormal(node->nodes[i], x, z, normal);
		}
	}

	if (count > 0)
	{
		return;
	}

	for (int i = 0; i < node->triangleCount; i++)
	{
		index = i * 3;
		vertexIndex = node->indices[index];
		vertex1[0] = m_heightMap[vertexIndex].x;
		vertex1[1] = m_heightMap[vertexIndex].y;
		vertex1[2] = m_heightMap[vertexIndex].z;

		index++;
		vertexIndex = node->indices[index];
		vertex2[0] = m_heightMap[vertexIndex].x;
		vertex2[1] = m_heightMap[vertexIndex].y;
		vertex2[2] = m_heightMap[vertexIndex].z;

		index++;
		vertexIndex = node->indices[index];
		vertex3[0] = m_heightMap[vertexIndex].x;
		vertex3[1] = m_heightMap[vertexIndex].y;
		vertex3[2] = m_heightMap[vertexIndex].z;

		foundHeight = CheckHeightOfTriangle(x, z, height, vertex1, vertex2, vertex3);

		if (foundHeight)
		{
			normal.x = m_heightMap[node->indices[index]].nx;
			normal.y = m_heightMap[node->indices[index]].ny;
			normal.z = m_heightMap[node->indices[index]].nz;
			return;
		}
	}
}

bool TerrainModel::CheckHeightOfTriangle(float x, float z, float& height, float v0[3], float v1[3], float v2[3])
{
	Vector3 startVector, directionVector, edge1, edge2, normal;
	Vector3 Q, e1, e2, e3, edgeNormal, temp;
	float magnitude, D, denominator, numerator, t, determinant;

	startVector.x = x;
	startVector.y = 0.0f;
	startVector.z = z;

	directionVector.x = 0.0f;
	directionVector.y = -1.0f;
	directionVector.z = 0.0f;

	edge1.x = v1[0] - v0[0];
	edge1.y = v1[1] - v0[1];
	edge1.z = v1[2] - v0[2];

	edge2.x = v2[0] - v0[0];
	edge2.y = v2[1] - v0[1];
	edge2.z = v2[2] - v0[2];

	normal.x = (edge1.y * edge2.z) - (edge1.z * edge2.y);
	normal.y = (edge1.z * edge2.x) - (edge1.x * edge2.z);
	normal.z = (edge1.x * edge2.y) - (edge1.y * edge2.x);

	normal.normalise();

	D = ((-normal.x * v0[0]) + (-normal.y * v0[1]) + (-normal.z * v0[2]));

	denominator = ((normal.x * directionVector.x) + (normal.y * directionVector.y) + (normal.z * directionVector.z));

	if (fabs(denominator) < 0.0001f)
	{
		return false;
	}

	numerator = -1.0f * (((normal.x * startVector.x) + (normal.y * startVector.y) + (normal.z * startVector.z)) + D);

	t = numerator / denominator;

	Q.x = startVector.x + (directionVector.x * t);
	Q.y = startVector.y + (directionVector.y * t);
	Q.z = startVector.z + (directionVector.z * t);

	e1.x = v1[0] - v0[0];
	e1.y = v1[1] - v0[1];
	e1.z = v1[2] - v0[2];

	e2.x = v2[0] - v1[0];
	e2.y = v2[1] - v1[1];
	e2.z = v2[2] - v1[2];

	e3.x = v0[0] - v2[0];
	e3.y = v0[1] - v2[1];
	e3.z = v0[2] - v2[2];

	edgeNormal.x = (e1.y * normal.z) - (e1.z * normal.y);
	edgeNormal.y = (e1.z * normal.x) - (e1.x * normal.z);
	edgeNormal.z = (e1.x * normal.y) - (e1.y * normal.x);

	temp.x = Q.x - v0[0];
	temp.y = Q.y - v0[1];
	temp.z = Q.z - v0[2];

	determinant = ((edgeNormal.x * temp.x) + (edgeNormal.y * temp.y) + (edgeNormal.z * temp.z));

	if (determinant > 0.001f)
	{
		return false;
	}

	edgeNormal.x = (e2.y * normal.z) - (e2.z * normal.y);
	edgeNormal.y = (e2.z * normal.x) - (e2.x * normal.z);
	edgeNormal.z = (e2.x * normal.y) - (e2.y * normal.x);

	temp.x = Q.x - v1[0];
	temp.y = Q.y - v1[1];
	temp.z = Q.z - v1[2];

	determinant = ((edgeNormal.x * temp.x) + (edgeNormal.y * temp.y) + (edgeNormal.z * temp.z));

	if (determinant > 0.001f)
	{
		return false;
	}

	edgeNormal.x = (e3.y * normal.z) - (e3.z * normal.y);
	edgeNormal.y = (e3.z * normal.x) - (e3.x * normal.z);
	edgeNormal.z = (e3.x * normal.y) - (e3.y * normal.x);

	temp.x = Q.x - v2[0];
	temp.y = Q.y - v2[1];
	temp.z = Q.z - v2[2];

	determinant = ((edgeNormal.x * temp.x) + (edgeNormal.y * temp.y) + (edgeNormal.z * temp.z));

	if (determinant > 0.001f)
	{
		return false;
	}

	height = Q.y + m_translate.y;

	return true;
}

void TerrainModel::CalculateNormals()
{
	int index1, index2, index3, index, count;
	Vector3 vertex1, vertex2, vertex3, vector1, vector2, sum, length;
	Vector3* normals;

	//Create a temporary array to hold normal vectors
	normals = new Vector3[(m_terrainHeight - 1) * (m_terrainWidth - 1)];

	for (int j = 0; j < (m_terrainHeight - 1); j++)
	{
		for (int i = 0; i < (m_terrainWidth - 1); i++)
		{
			index1 = ((j + 1) * m_terrainWidth) + i;
			index2 = ((j + 1) * m_terrainWidth) + (i + 1);
			index3 = (j * m_terrainWidth) + i;

			vertex1.x = m_heightMap[index1].x;
			vertex1.y = m_heightMap[index1].y;
			vertex1.z = m_heightMap[index1].z;

			vertex2.x = m_heightMap[index2].x;
			vertex2.y = m_heightMap[index2].y;
			vertex2.z = m_heightMap[index2].z;

			vertex3.x = m_heightMap[index3].x;
			vertex3.y = m_heightMap[index3].y;
			vertex3.z = m_heightMap[index3].z;

			vector1 = vertex1 - vertex3;
			vector2 = vertex3 - vertex2;

			index = (j * (m_terrainWidth - 1)) + i;

			normals[index] = vector1.cross(vector2);

			normals[index].normalise();
		}
	}

	for (int j = 0; j < m_terrainHeight; j++)
	{
		for (int i = 0; i < m_terrainWidth; i++)
		{
			sum.x = 0.0f;
			sum.y = 0.0f;
			sum.z = 0.0f;

			//bottom left
			if (((i - 1) >= 0) && ((j - 1) >= 0))
			{
				index = ((j - 1) * (m_terrainWidth - 1)) + (i - 1);
				sum += normals[index];
			}

			//bottom right
			if ((i < (m_terrainWidth - 1)) && ((j - 1) >= 0))
			{
				index = ((j - 1) * (m_terrainWidth - 1)) + i;
				sum += normals[index];
			}

			//upper left
			if (((i - 1) >= 0) && (j < (m_terrainHeight - 1)))
			{
				index = (j*(m_terrainWidth - 1)) + i;
				sum += normals[index];
			}

			//upper right
			if ((i < (m_terrainWidth - 1)) && (j < (m_terrainHeight - 1)))
			{
				index = (j * (m_terrainWidth - 1)) + i;
				sum += normals[index];
			}

			index = (j*m_terrainWidth) + i;
			sum.normalise();
			m_heightMap[index].nx = sum.x;
			m_heightMap[index].ny = sum.y;
			m_heightMap[index].nz = sum.z;
		}
	}

	delete[] normals;
}