#ifndef _TERRAINMODEL_H_
#define _TERRAINMODEL_H_

#include <thread>
#include <future>
#include <mutex>
#include <iostream>
#include <list>
#include <chrono>
#include <atomic>
#include <sstream>
#include <vector>
#include <fstream>
#include <ctime>
#include <algorithm>


#include "Model.h"
#include "ThreadChannel.h"

const int MAX_TRIANGLES = 10000;
const int THREAD_COUNT = 8;
const long long MAX_CHECK_TIME = 1000;
const long long MAX_MANIPULATION_TIME = 4;
const long long MAX_NORMAL_TIME = 1000;

using std::thread;
using std::promise;
using std::mutex;
using std::list;
using std::atomic_bool;
using std::vector;

using std::chrono::duration_cast;
using std::chrono::milliseconds;
typedef std::chrono::steady_clock the_clock;


enum class AttenuationType
{
	CONSTANT, LINEAR, QUADRATIC
};

class TerrainModel : public Model
{
private:
	struct NodeType
	{
		float positionX, positionZ, width;
		int triangleCount;
		unsigned long* indices;
		NodeType* nodes[4];
	};

	struct TriangleThreadArgs
	{
		int start_index, end_index;
		float positionX, positionZ, width;
		int* count;
	};

	struct NodeThreadArgs
	{
		int start_index, end_index;
		float positionX, positionZ, width;
		NodeType* node;
		int* index;
	};

	struct TaskListEntry
	{
		unsigned long index;
		float distance;
		float manipulation_intensity;
		float attenuation_amount;
		AttenuationType attenuation_type;
		int frames_since_creation;
		bool operator<(TaskListEntry& other)
		{
			if (frames_since_creation < other.frames_since_creation)
			{
				return true;
			}
			else if (frames_since_creation > other.frames_since_creation)
			{
				return false;
			}
			else
			{
				if (index < other.index)
				{
					return true;
				}
				else if (index > other.index)
				{
					return false;
				}
				else
				{
					return false;
				}
			}
		}

		bool operator==(TaskListEntry& other)
		{
			return (frames_since_creation == other.frames_since_creation && index == other.index);
		}
	};

	struct CheckVertexArgs
	{
		Vector3 origin_position;
		float radius;
		float manipulation_intensity;
		float attenuation_amount;
		AttenuationType attenuation_type;
		NodeType* node;
		int start_index, end_index;
	};

	struct DataSnapshotEntry
	{
		long long task_list_time;
		vector<int> tasks_processed;
	};

public:
	bool Initialize(char* filename);
	void Frame();
	void Render();
	void Shutdown();
	bool GetHeightAtPosition(float x, float z, float& height);
	bool GetNormalAtPosition(float x, float z, Vector3& normal);
	void ManipulateTerrain(Vector3 manipulation_origin, float manipulation_radius, float manipulation_intensity, float attenuation_amount, AttenuationType attenuation_type);
	void WriteSnapshot(char* filename);

private:
	bool LoadHeightMap(char* filename);
	bool LoadRawHeightMap(char* filename);
	void InitializeQuadTree();
	void CalculateMeshDimensions(int vertexCount, float& centerX, float& centerZ, float& meshWidth);
	void CreateTreeNode(NodeType* parentNode, float centerX, float centerZ, float width);
	void CalculateNodeThreaded(NodeThreadArgs args);
	void RenderNode(NodeType* node);
	void ReleaseNode(NodeType* node);
	int CountTriangles(float positionX, float positionZ, float width);
	bool IsTriangleContained(int triangleNum, float positionX, float positionZ, float width);
	void IsTriangleContainedThreaded(TriangleThreadArgs args);
	void FindNodeHeight(NodeType* node, float x, float z, float& height);
	void FindNodeNormal(NodeType* node, float x, float z, Vector3& normal);
	void FindNode(NodeType* parent_node, float radius, float x, float z);
	bool CheckHeightOfTriangle(float x, float z, float& height, float v0[3], float v1[3], float v2[3]);
	void SetTerrainCoordinates();
	bool BuildTerrainModel();
	void CheckVertices(CheckVertexArgs args);
	void ManipulateVertices(channel<TaskListEntry>* task_channel, atomic_bool* elapsed, int* task_count, int thread_num);
	void CalculateNormals();
	bool SameData(const TaskListEntry& task1, const TaskListEntry& task2);
	void InitializeTerrainThreading();

private:
	int m_terrainWidth, m_terrainHeight;
	int m_triangleCount;
	NodeType* m_parentNode;
	VertexType* m_heightMap;
	mutex count_mutex;
	float m_meshMinX, m_meshMaxX, m_meshMinZ, m_meshMaxZ;
	vector<TaskListEntry> m_taskList;
	thread m_threadList[THREAD_COUNT];
	mutex task_list_mutex;
	mutex task_count_mutex;
	channel<TaskListEntry> m_taskChannel;
	vector<DataSnapshotEntry> m_snapshotList;
	vector<NodeType*> m_nodeList;
	atomic_bool elapsed;
	int task_count;
};

#endif